/***************************************************************************
*
*   Module:             diags
*
*   Filename:           sysevent.c
*
*   Author:             Brian Tran
*
*   Created:            09/03/2002
*
*   Description:
*
*   This device is read-only from user perspective. The bloking read returns when there is an interrput on system events
*   such as fan faults, over temperature, power supply failure. A user library is needed to find out what exactly what when wrong.
*   This device only returns the one-byte cause code of the event.  
*
****************************************************************************
*                 Source Control System Information
*
*    $Id: 120sc_sysevent.c,v 1.17 2005/05/03 01:38:14 brian Exp $
*
****************************************************************************
*                       Modification History
*
*    $Log: 120sc_sysevent.c,v $
*    Revision 1.17  2005/05/03 01:38:14  brian
*    Bugzilla number:
*    Reset the i2c switch and retry in case of a switch error
*
*    Revision 1.16  2005/05/02 01:12:13  brian
*    Bugzilla number:
*    Add ECC reporting in 120
*
*    Revision 1.15  2005/04/19 00:47:05  brian
*    Bugzilla number:
*    Support run-time fru error for 120
*
*    Revision 1.14  2005/01/07 12:04:37  vijayendra
*    Bugzilla number:5824
*    Fixed the reversed post error storage array
*
*    Revision 1.13  2004/11/30 01:58:41  brian
*    Bugzilla number: 5350
*    Keep generating fan interrupts once a fan fails to support shut down without working fans
*
*    Revision 1.12  2004/10/06 07:05:19  brian
*    Bugzilla number:
*    Support for POST result in 120
*
*    Revision 1.11  2004/09/20 20:46:52  brian
*    Bugzilla number:
*    Add reset reason for 120. No longer support voltage/current fault recovery.
*
*    Revision 1.10  2004/08/03 20:31:18  ted
*    Add support for Anafa2 as a level-triggered acvive low interrupt.
*    Clean up some formatting.
*
*    Revision 1.9  2004/04/20 01:59:05  brian
*    Bugzilla number:
*    Copy the hang problem workaround from 2.0 here
*
*    Revision 1.8  2004/02/21 02:15:02  brian
*    Bugzilla number:
*    Support hardware err sync interrupt
*
*    Revision 1.7  2004/02/20 04:42:28  brian
*    Bugzilla number:
*    Add hardware error interrupt
*
*    Revision 1.6  2004/02/12 04:45:45  brian
*    Bugzilla number:
*    Support system event in 120
*
*    Revision 1.5  2004/01/12 22:49:10  mshannon
*    Move this code (which changes led state to solid green) out to user space.
*
*    Revision 1.4  2003/12/18 05:12:29  brian
*    Bugzilla number:
*    Add i2c debug hook
*
*    Revision 1.3  2003/12/10 00:58:41  mshannon
*    ppcboot had started front panel leds blinking green, now kernel
*    turns them on solid green.
*
*    Revision 1.2  2003/10/30 23:08:31  brian
*    Bugzilla number:
*    Support seti2croute in 120 kernel
*
*    Revision 1.1  2003/10/29 01:16:00  brian
*    Bugzilla number:
*    For 120
*
*
***************************************************************************
*                       (c) COPYRIGHT 2000, 2001, 2002
*                           TopSpin Communications, Inc.
*                           Palo Alto, CA
**************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id: 120sc_sysevent.c,v 1.17 2005/05/03 01:38:14 brian Exp $";
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include "120sc_sysevent.h"
#include <linux/i2c-errcode.h>
#include "ds1672_rtc.h"
#include "dcr.h"
#include "ecc.h"
#include <system_interrupt_code.h>
#include <fpga.h>

extern int global_i2c_debug;
extern int i2c_controller_fast_mode;

extern int phy_iic_xfer(void *vaddr, unsigned short dev_addr, int access, char *buf, int count, int controller);

int sysevent_init_module(void);
void sysevent_cleanup_module(void);
int seti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller);


/* Function Prototypes */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int sysevent_poll(struct file *file, poll_table *wait);
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

/* forward declarations */
void init_ts_watchdog(void);
void ts120_anafa_reset(void);
void ts120_disable_dma(void);




/* File Operation Table */
static struct file_operations char_fops = {
    owner: THIS_MODULE,
    read: sysevent_read,
    write: sysevent_write,
    ioctl: sysevent_ioctl,
    poll: sysevent_poll,
};

static wait_queue_head_t sysevent_waitq;
static wait_queue_head_t delay_waitq;



static char *ts120_reset_reason;
static volatile int fan_fault_status;
static volatile int fan_fault_int_active;
static volatile int fan_recover_int_active;
static volatile int smm_fault_int_active;
static volatile int smm_fault_recover_int_active;
static volatile int fpga_int_active;
static volatile int ps_fail_int_active;
static volatile int optical_fault_int_active;
static volatile int smm_healthy_int_active;
static volatile int smm_healthy_recover_int_active;
static volatile int hardware_err_sync_int_active;

#define MAX_120_FRU		12
static volatile int err_fru_id[MAX_120_FRU];
static unsigned char post_err[MAX_120_FRU][POST_RESULT_LEN];
static int fru_err[MAX_120_FRU];


static int timeout = 20;

static void *fpga_base_addr;
static volatile void *gpio_base_addr;


/* interrupt handler */

static void fan_fault_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC1_POLARITY_REG);
    if (dcr_val & TS120_FAN_INT_MASK) {		/* the bit is 1, rising edge - recovery event */
        fan_recover_int_active = 1; 
    }
    else {
        fan_fault_int_active = 1;
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS120_FAN_INT_MASK;
    mtdcr(UIC1_POLARITY_REG, dcr_val);

    wake_up_interruptible(&sysevent_waitq);
}

static void smm_fault_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
#if 0
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    if (dcr_val & TS120_SMM_FAULT_INT_MASK) {       /* the bit is 1, rising edge - recovery event */
        smm_fault_recover_int_active = 1;
    }
    else {
        smm_fault_int_active = 1;
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS120_SMM_FAULT_INT_MASK;
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    wake_up_interruptible(&sysevent_waitq);
#endif
}

static void fpga_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    fpga_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
}

static void ps_fail_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    ps_fail_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
}   

static void optical_fault_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    optical_fault_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
}

static void smm_healthy_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
#if 0
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    if (dcr_val & TS120_SMM_HEALTHY_INT_MASK) {       /* the bit is 1, rising edge - recovery event */
        smm_healthy_recover_int_active = 1;
    }
    else {
        smm_healthy_int_active = 1;
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS120_SMM_HEALTHY_INT_MASK;
    mtdcr(UIC0_POLARITY_REG, dcr_val);
#endif

    smm_healthy_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);

}

int read_reset_reason(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
    if (offset > 0) {
        return 0;
    }

    len = sprintf(buf, ts120_reset_reason);
    buf[len] = '\n';
    len ++;
    *eof = 1;

    return len;
}



int sysevent_init_module(void)
{
    int dcr_val;
    unsigned char fpga_status;
    char *sram_sig;
    int fru;
    int i;

    /* init global vars */
    fan_fault_status = 0; 
    fan_fault_int_active = 0; 
    fan_recover_int_active = 0; 
    smm_fault_int_active = 0;
    smm_fault_recover_int_active = 0;
    fpga_int_active = 0;
    ps_fail_int_active = 0;
    optical_fault_int_active = 0;
    smm_healthy_int_active = 0;
    smm_healthy_recover_int_active = 0;
    hardware_err_sync_int_active = 0;

    for (fru = 0; fru < MAX_120_FRU; fru ++) {
       err_fru_id[fru] =  0;
       fru_err[fru] = 0;
       for (i = 0; i < POST_RESULT_LEN; i ++) {
           post_err[fru][i] = 0;
       }
    }

    /* register char device */
    if (register_chrdev(SYSEVENT_MAJOR, "syseventmodule", &char_fops)) {
        printk(KERN_ERR "Unable to register syseventmodule\n");
        return -EIO;
    }

    create_proc_read_entry("resetreason", 0, NULL, read_reset_reason, NULL);

    sram_sig = (char *)PPC440_SRAM_VIRT_BASE;

    /* inspect the sram for reset signature */
    sram_sig[RESET_SIG_LEN] = 0;

    if (!strcmp(sram_sig, SOFTWARE_RESET)) {
        ts120_reset_reason = SOFTWARE_RESET;
    }
    else if (!strcmp(sram_sig, WATCHDOG_RESET)) {
        ts120_reset_reason = WATCHDOG_RESET;
    }
    else {
        ts120_reset_reason = POWER_ON_RESET;
    }

    strcpy(sram_sig, WATCHDOG_RESET);


    fpga_base_addr = ioremap_nocache(FPGA_REGADDR_BASE, 0x200); 
    gpio_base_addr = (volatile void *)ioremap_nocache(GPIO_BASE_REGADDR, 0x1000);

    /* make sure gpio 11: tri-state disable, no open drain */
        {
            volatile unsigned int data;
            /* disable tri-state - make sure tri-state bit == 1 */
            data = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_TRISTATE_REGADDR)));
            data |= GPIO_11_MASK | GPIO_9_MASK;
            writel(__cpu_to_le32(data), (volatile void *)((unsigned int)gpio_base_addr + GPIO_TRISTATE_REGADDR));

            /* no open drain - make sure open drain bit == 0 */
            data = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OPENDRAIN_REGADDR)));
            data &= ~(GPIO_11_MASK | GPIO_9_MASK);
            writel(__cpu_to_le32(data), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OPENDRAIN_REGADDR));
        }


    /* init wait queue */
    init_waitqueue_head(&sysevent_waitq);
    init_waitqueue_head(&delay_waitq);

    /* configure IRQ 1, 2, 4, 5, 6, 10 to be trigger type, falling edge */
    dcr_val = mfdcr(UIC0_TRIGGER_REG);
    dcr_val |= TS120_UIC0_EDGE_INT_MASK;
    dcr_val &= ~TS120_UIC0_LEVEL_INT_MASK;
    mtdcr(UIC0_TRIGGER_REG, dcr_val);

    /* for IRQ10 bit 21 */
    dcr_val = mfdcr(UIC1_TRIGGER_REG);
    dcr_val |= TS120_FAN_INT_MASK;
    mtdcr(UIC1_TRIGGER_REG, dcr_val);

    /* All of our interrupts are active-low */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    dcr_val &= (~TS120_UIC0_INT_MASK);
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    dcr_val = mfdcr(UIC1_POLARITY_REG);
    dcr_val &= (~TS120_FAN_INT_MASK);
    mtdcr(UIC1_POLARITY_REG, dcr_val);


    /* clear int status */
    dcr_val = mfdcr(UIC0_STATUS_REG);
    dcr_val |= TS120_UIC0_INT_MASK;
    mtdcr(UIC0_STATUS_REG, dcr_val);

    dcr_val = mfdcr(UIC1_STATUS_REG);
    dcr_val |= TS120_FAN_INT_MASK;
    mtdcr(UIC1_STATUS_REG, dcr_val);


    /* request interrupt */
    if (request_irq(FAN_FAULT_IRQ_NUM, fan_fault_int_handler, SA_INTERRUPT, "Fan Fault Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", FAN_FAULT_IRQ_NUM);
        return -1;
    }

    if (request_irq(SMM_FAULT_IRQ_NUM, smm_fault_int_handler, SA_INTERRUPT, "Summit Fault Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", SMM_FAULT_IRQ_NUM);
        return -1;
    }

    if (request_irq(FPGA_IRQ_NUM, fpga_int_handler, SA_INTERRUPT, "FPGA Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", FPGA_IRQ_NUM);
        return -1;
    }

    if (request_irq(PS_FAIL_IRQ_NUM, ps_fail_int_handler, SA_INTERRUPT, "Power Supply Fail Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", PS_FAIL_IRQ_NUM);
        return -1;
    }

    if (request_irq(OPTICAL_FAULT_IRQ_NUM, optical_fault_int_handler, SA_INTERRUPT, "Optical Fault Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", OPTICAL_FAULT_IRQ_NUM);
        return -1;
    }

    if (request_irq(SMM_HEALTHY_IRQ_NUM, smm_healthy_int_handler, SA_INTERRUPT, "Summit Healthy Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", SMM_HEALTHY_IRQ_NUM);
        return -1;
    }




    /* clear fpga change register so that any new interrupt will be seen */
    writeb(0xff, (void *) ((int) fpga_base_addr + SC_120_FPGA_PS_THERM_1WIRE_CHANGE_REG_OFFSET));
    writeb(0xff, (void *) ((int) fpga_base_addr + SC_120_FPGA_MEZZINT_CHANGE_REG_OFFSET));

    /* enable interrupt on changes */
    writeb(SC_120_FPGA_DEFAULT_INT_MASK, (void *) ((int) fpga_base_addr + SC_120_FPGA_PS_THERM_1WIRE_ENABLE_REG_OFFSET));
    writeb(SC_120_FPGA_TOP_MEZZINT_MASK | SC_120_FPGA_BOTTOM_MEZZINT_MASK, (void *) ((int) fpga_base_addr + SC_120_FPGA_MEZZINT_ENABLE_REG_OFFSET));

    /* in case something bad has occurred before this routine is called, let's read fpga status register */
    fpga_status = readb((void *) ((int) fpga_base_addr + SC_120_FPGA_PS_THERM_1WIRE_STATUS_REG_OFFSET));
    if (((fpga_status & SC_120_FPGA_PS1_ACFAIL_MASK) != 0) ||
        ((fpga_status & SC_120_FPGA_PS2_ACFAIL_MASK) != 0) ||
        ((fpga_status & SC_120_FPGA_PS1_DCFAIL_MASK) != 0) ||
        ((fpga_status & SC_120_FPGA_PS2_DCFAIL_MASK) != 0) ||
        ((fpga_status & SC_120_FPGA_PS1_PRESENT_MASK) != 0) ||
        ((fpga_status & SC_120_FPGA_PS2_PRESENT_MASK) != 0)) {
        fpga_int_active = 1;
    }

    fpga_status = readb((void *) ((int) fpga_base_addr + SC_120_FPGA_MEZZINT_STATUS_REG_OFFSET));
    if (((fpga_status & SC_120_FPGA_TOP_MEZZINT_MASK) != 0) ||
        ((fpga_status & SC_120_FPGA_BOTTOM_MEZZINT_MASK) != 0)) {
        fpga_int_active = 1;
    }

    /* add ECC */
    ecc_init();
        
    return 0;
}


void sysevent_cleanup_module(void)
{

    unregister_chrdev(SYSEVENT_MAJOR, "syseventmodule");
    remove_proc_entry("resetreason", NULL);
    iounmap(fpga_base_addr);
    iounmap((void *)gpio_base_addr);
    free_irq(FAN_FAULT_IRQ_NUM, NULL);
    free_irq(SMM_FAULT_IRQ_NUM, NULL);
    free_irq(FPGA_IRQ_NUM, NULL);
    free_irq(PS_FAIL_IRQ_NUM, NULL);
    free_irq(OPTICAL_FAULT_IRQ_NUM, NULL);
    free_irq(SMM_HEALTHY_IRQ_NUM, NULL);

    /* ECC */
    ecc_cleanup();
}




/* Read routine */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
    int interrupt_code;
    int delay;
    unsigned char fpga_change;
    int fru;
    int len;
    int minor;
    struct inode *inode = file->f_dentry->d_inode;

    minor = MINOR(inode->i_rdev);

    if (minor == SYSEVENT_ECC_MINOR) {
        return ecc_read(file, buf, count, offset);
    }

    if (minor != 0) {
        if (count > POST_RESULT_LEN) {
            len = POST_RESULT_LEN;
        }
        else {
            len = count;
        }
        copy_to_user(buf, &post_err[minor][0], len);

        return len;
    }


    /* if here, default minor == 0 */

    if ((fan_fault_int_active == 0) && (fan_recover_int_active == 0) &&
        (smm_fault_int_active == 0) && (smm_fault_recover_int_active == 0) &&
        (fpga_int_active == 0) && (ps_fail_int_active == 0) &&
        (optical_fault_int_active == 0) && (smm_healthy_int_active == 0) &&
        (hardware_err_sync_int_active == 0) &&
        (smm_healthy_recover_int_active == 0)) {
        interruptible_sleep_on_timeout(&sysevent_waitq, timeout * HZ);
    }

    if (hardware_err_sync_int_active == 1) {
        interrupt_code = HARDWARE_ERR_SYNC_INT_CODE;
        /* scan to see which fru has errors */
        for (fru = 0; fru < MAX_120_FRU; fru ++) {
            if (err_fru_id[fru] != 0) {
                interrupt_code |= err_fru_id[fru] << SYSTEM_INT_SHIFT;
                err_fru_id[fru] =  0;
                hardware_err_sync_int_active = 0;
                break;
            }
        }
        /* scan one more time to check any other pending errors */
        for (fru = 0; fru < MAX_120_FRU; fru ++) {
            if (err_fru_id[fru] != 0) {
                hardware_err_sync_int_active = 1;
                break;
            }
        }
        delay = 0;
        goto sysevent_read_return;
    }


    if (fan_fault_int_active == 1) {
        fan_fault_status = 1;
        interrupt_code = FAN_INT_CODE;
        fan_fault_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (fan_recover_int_active == 1) {
        fan_fault_status = 0;
        interrupt_code = FAN_RECOVER_INT_CODE;
        fan_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }


    if (smm_fault_int_active == 1) {
        interrupt_code = SMM_FAULT_INT_CODE;
        smm_fault_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_fault_recover_int_active == 1) {   
        interrupt_code = SMM_FAULT_RECOVER_INT_CODE;
        smm_fault_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }


    if (fpga_int_active == 1) {     
        interrupt_code = FPGA_INT_CODE;
        fpga_int_active = 0;
        delay = 1;
        goto sysevent_read_return;
    }

    if (ps_fail_int_active == 1) {    
        interrupt_code = PS_FAIL_INT_CODE;
        ps_fail_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (optical_fault_int_active == 1) {
        interrupt_code = OPTICAL_FAULT_INT_CODE;
        optical_fault_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_healthy_int_active == 1) {
        interrupt_code = SMM_HEALTHY_INT_CODE;
        smm_healthy_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_healthy_recover_int_active == 1) {
        interrupt_code = SMM_HEALTHY_RECOVER_INT_CODE;
        smm_healthy_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }



    fpga_change = readb((void *) ((int) fpga_base_addr + SC_120_FPGA_PS_THERM_1WIRE_CHANGE_REG_OFFSET));
    if ((fpga_change & SC_120_FPGA_DEFAULT_INT_MASK) != 0) {
        interrupt_code = FPGA_INT_CODE;
        delay = 1;
        goto sysevent_read_return;
    }

    fpga_change = readb((void *) ((int) fpga_base_addr + SC_120_FPGA_MEZZINT_CHANGE_REG_OFFSET));
    if (((fpga_change & SC_120_FPGA_TOP_MEZZINT_MASK) != 0) ||
        ((fpga_change & SC_120_FPGA_BOTTOM_MEZZINT_MASK) != 0)) {
        interrupt_code = FPGA_INT_CODE;
        delay = 1;
        goto sysevent_read_return;
    }

    /* once fan fault status == 1, keep generating the fan int code to poll for another fan fault or recovery */ 
    if (fan_fault_status == 1) {
        interrupt_code = FAN_INT_CODE;
        delay = 0;
        goto sysevent_read_return;
    }

    /* return with interrupt code == 0 */
    interrupt_code = 0;
    delay = 0;


sysevent_read_return:

    /* delay before returning */
    interruptible_sleep_on_timeout(&delay_waitq, delay * HZ);
    copy_to_user(buf, &interrupt_code, sizeof(int));
    return sizeof(int);

}


/* Write routine
 * -------------
 * Users want to clear a bit mask in the interrupt code. The value should have the bit zero, all other bits are 1
 */
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset)
{
    int value;
    int minor;
    int fruid;
    int len;
    struct inode *inode = file->f_dentry->d_inode;

    minor = MINOR(inode->i_rdev);

    if (minor != 0) {
        if (count > POST_RESULT_LEN) {
            len = POST_RESULT_LEN;
        }
        else {
            len = count;
        }
        copy_from_user(&post_err[minor][0], buf, len);

        return len;
    }

    /* if here default minor == 0 */
    copy_from_user(&value, buf, sizeof(int));

    if ((value & I2C_DEBUG_CODE) == I2C_DEBUG_CODE) {
        global_i2c_debug = value & 0xff;
    }

    if ((value & I2C_CLOCK_CODE) == I2C_CLOCK_CODE) {
        i2c_controller_fast_mode = value & 0xff;
    }

    if ((value & FRU_ERR_CODE) == FRU_ERR_CODE) {
        fruid = value & SYSEVENTWRITE_FRUID_MASK;
        if ((fruid >= 0) && (fruid < MAX_120_FRU)) { 
            hardware_err_sync_int_active = 1;
            err_fru_id[fruid] = value & 0xffff;
            wake_up_interruptible(&sysevent_waitq);
        }
    }


    return sizeof(int);
}

/* Ioctl routine 
 * -------------
 * Use this for run-time error support.
 */
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    int command;
    int fruid;

    fruid = cmd & SYSEVENTWRITE_FRUID_MASK;
    command = cmd & (~SYSEVENTWRITE_FRUID_MASK); 

    switch(command) {
        case TS120_FRUERR_READ:
            return (copy_to_user((int *)arg, &fru_err[fruid], sizeof(int))) ? -EFAULT: 0;
            break;
        case TS120_FRUERR_WRITE:
            fru_err[fruid] = arg;
            break;
        case TS120_ECC_COUNT_READ:
            break;
        case TS120_ECC_CLEAR:
            break;
        default:
            break;
    }

    return 0;
}


static unsigned int sysevent_poll(struct file *file, poll_table *wait)
{
    return 0;
}



/* seti2croute
 * -----------
 * In 120, i2c bus 1 has a switch that allows access to Summit (channel 0), PS A (channel 1), PS B (channel 2), and Anafa Seeproms (channel 3).
 * The mux select parameter of this function has the value of the channel number, which selects one of these channels.
 */   
int seti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller)
{
    unsigned char data;
    unsigned char misc_control_reg0;

    if (muxselect > 3) {
        return 0;
    }

    if (controller == 0) {      /* on 120, controller 0 is NOT connected to the i2c switch */
        return 0;
    }

    data = 1 << muxselect;

    if (iic_xfer(vaddr, (0xff << 8) | LOCAL_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller) != 1) {
        /* reset the switch and retry */
        misc_control_reg0 = readb((void *) ((int) fpga_base_addr + SC_120_FPGA_MISC_CONTROL_REG0_OFFSET));
        misc_control_reg0 |= SC_120_FPGA_I2CSWITCH_RESET_MASK;
        writeb(misc_control_reg0, (void *) ((int) fpga_base_addr + SC_120_FPGA_MISC_CONTROL_REG0_OFFSET));
        udelay(1);
        misc_control_reg0 &= ~SC_120_FPGA_I2CSWITCH_RESET_MASK;
        writeb(misc_control_reg0, (void *) ((int) fpga_base_addr + SC_120_FPGA_MISC_CONTROL_REG0_OFFSET));
        udelay(1);
        if (iic_xfer(vaddr, (0xff << 8) | LOCAL_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller) != 1) {
            return IIC_ERR_LOCALSWITCH_ERR;
        }
    }

    return 0;

}


/* init_ts_watchdog
 * ------------
 * Start the hardware watchdog.
 */
void init_ts_watchdog()
{
    volatile unsigned int gpio_output;

    gpio_output = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR)));

    /* write 1 */
    gpio_output |= GPIO_11_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    /* write 0 */
    gpio_output &= ~GPIO_11_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    return;
}

    
/* ts120_anafa_reset
 * ------------
 */
void ts120_anafa_reset()
{
    volatile unsigned int gpio_output;

    gpio_output = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR)));

    /* write 0 */
    gpio_output &= ~GPIO_9_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));
 
    return;
}


/* ts120_disable_dma
 * -----------------
 */
void ts120_disable_dma()
{   
    writeb(0x1, (void *) ((int) fpga_base_addr + 0x5C));
}






