/***************************************************************************
*
*   Module:             diags
*
*   Filename:           sysevent.c
*
*   Author:             Brian Tran
*
*   Created:            09/03/2002
*
*   Description:
*
*   This device is read-only from user perspective. The bloking read returns when there is an interrput on system events
*   such as fan faults, over temperature, power supply failure. A user library is needed to find out what exactly what when wrong.
*   This device only returns the one-byte cause code of the event.  
*
****************************************************************************
*                 Source Control System Information
*
*    $Id: 270sc_sysevent.c,v 1.51 2005/04/30 03:44:43 brian Exp $
*
****************************************************************************
*                       Modification History
*
*    $Log: 270sc_sysevent.c,v $
*    Revision 1.51  2005/04/30 03:44:43  brian
*    Bugzilla number:
*    Add variable to save runtime error events
*
*    Revision 1.50  2005/04/27 09:33:22  brian
*    Bugzilla number:
*    Support runtime error reporting
*
*    Revision 1.49  2005/03/16 22:42:32  brian
*    Bugzilla number:
*    Increase max FRU number in 270 to include optical modules
*
*    Revision 1.48  2005/02/24 00:49:02  brian
*    Bugzilla number: 6026
*    No longer use err out for PICInit
*
*    Revision 1.47  2005/02/18 01:11:48  brian
*    Bugzilla number: 6026
*    Allow mgmt mux sharing during PIC update
*
*    Revision 1.46  2005/01/07 12:02:51  vijayendra
*    Bugzilla number: 5802,5804
*    Fixed the reversed post error storage array
*
*    Revision 1.45  2005/01/06 01:31:30  brian
*    Bugzilla number:
*    Support for optical LIMs
*
*    Revision 1.44  2005/01/05 02:02:26  brian
*    Bugzilla number:
*    Add additional muxselect byte to support optical LIM in 270
*
*    Revision 1.43  2004/11/19 02:30:30  brian
*    Bugzilla number:
*    Increase the polling interval
*
*    Revision 1.42  2004/11/18 01:43:27  brian
*    Bugzilla number:
*    Slower i2c clock back to 100KHz as we are getting more i2c access errors.
*
*    Revision 1.41  2004/11/17 00:15:18  brian
*    Bugzilla number:
*    When a core accesses thr other mgmt card, check for error out asserted by the other core in case of PICInit
*
*    Revision 1.40  2004/11/12 03:29:21  brian
*    Bugzilla number:
*    Dynamically change i2c clock. Use 100 KHz to access Summit device. All other devices use 400 KHz
*
*    Revision 1.39  2004/11/10 22:22:47  brian
*    Bugzilla number:
*    Recent changes to allow PICInit from both cores caused some problem (slowness). Removed thoses changes for now.
*
*    Revision 1.38  2004/10/26 20:52:47  brian
*    Bugzilla number:
*    Allow mgmt mux sharing while in PIC update
*
*    Revision 1.37  2004/09/24 00:33:37  brian
*    Bugzilla number:
*    Support getting POST results
*
*    Revision 1.36  2004/09/15 02:21:53  brian
*    Bugzilla number:
*    Add intr code to poll slave core current/voltage status
*
*    Revision 1.35  2004/08/31 23:10:25  brian
*    Bugzilla number:
*    Do not check fru id 0 for 270
*
*    Revision 1.34  2004/08/31 01:43:47  brian
*    Bugzilla number:
*    Disallow mgmt access from standby core when master is doing PIC access
*
*    Revision 1.33  2004/08/19 22:01:12  brian
*    Bugzilla number:
*    Change mux timeout to 2 seconds
*
*    Revision 1.32  2004/08/19 21:56:41  brian
*    Bugzilla number:
*    Fixed a bug in mgmt mux override mode
*
*    Revision 1.31  2004/08/17 23:07:57  brian
*    Bugzilla number:
*    Toggle node intr line instead of aseerting it when smm healthy inr is active
*
*    Revision 1.30  2004/08/07 02:06:33  brian
*    Bugzilla number:
*    Wake the wait queue on SMM Healthy intr
*
*    Revision 1.29  2004/08/06 02:05:18  brian
*    Bugzilla number:
*    No longer support temp intr. No longer support current and voltage recovery.
*    Allow mgmt mux override when timeout, and restore mgmt mux arbitration when mux is available again.
*
*    Revision 1.28  2004/08/03 21:14:05  brian
*    Bugzilla number:
*    Increase waiting time fo mgmt mux some more
*
*    Revision 1.27  2004/07/31 01:54:11  brian
*    Bugzilla number:
*    Extend wait time to compete for mgmt mux.
*
*    Revision 1.26  2004/07/27 17:36:14  brian
*    Bugzilla number:
*    Make fru error global so that i2c slave can write to it
*
*    Revision 1.25  2004/07/16 02:06:57  brian
*    Bugzilla number:
*    Enable node err_in intr
*
*    Revision 1.24  2004/07/15 00:31:03  uid1017
*    Bugzilla number:
*    Allow each FRU to send error code concurrently
*
*    Revision 1.23  2004/07/13 19:00:50  brian
*    Bugzilla number:
*    Cleanup intr handlers
*
*    Revision 1.22  2004/07/01 00:47:15  brian
*    Bugzilla number:
*    Now fpga_base_addr is a global var. Also allow to auto select mgmt card if there is only one present.
*
*    Revision 1.21  2004/06/29 01:59:49  brian
*    Bugzilla number:
*    Support polling in 270
*
*    Revision 1.20  2004/06/22 21:00:28  brian
*    Bugzilla number:
*    Add unseti2croute. Also add code to restrict i2c accesses from non-master cards. Add code to allow tow cores to compete for them same mgmt mux.
*
*    Revision 1.19  2004/06/16 22:21:04  brian
*    Bugzilla number:
*    Not allow 270 cores to share one mgmt card to avoid contention
*
*    Revision 1.18  2004/05/13 08:34:52  brian
*    Bugzilla number:
*    Support for getting reset reason in 270
*
*    Revision 1.17  2004/05/11 07:51:29  brian
*    Bugzilla number:
*    Support for changing i2c speed on the fly
*
*    Revision 1.16  2004/05/06 21:47:08  brian
*    Bugzilla number:
*    Check the masterchange bit first before waiting on a queue
*
*    Revision 1.15  2004/05/06 00:12:26  brian
*    Bugzilla number:
*    Support for mastership change intr in 270
*
*    Revision 1.14  2004/05/05 06:39:54  brian
*    Bugzilla number:
*    Route smm intr from a node to fpga intr in a core
*
*    Revision 1.13  2004/04/23 21:37:57  brian
*    Bugzilla number:
*    Support for i2croute with single mgmt card in any slot
*
*    Revision 1.12  2004/04/20 02:16:26  brian
*    Bugzilla number:
*    Add the reset hang fix for 270
*
*    Revision 1.11  2004/04/02 03:04:00  brian
*    Bugzilla number:
*    Now turn off all intr and let the master turn it on later in system/hw_lib/init.c
*
*    Revision 1.10  2004/04/01 02:40:29  brian
*    Bugzilla number:
*    Disable unused interrupts
*
*    Revision 1.9  2004/03/24 03:56:22  brian
*    Bugzilla number:
*    Initial support for system event handling
*
*    Revision 1.8  2004/01/13 23:53:51  sarah
*    put back previous changes, fix header file includes
*
*    Revision 1.6  2003/12/12 15:56:45  sarah
*    init intrs
*
*    Revision 1.5  2003/11/12 21:25:23  brian
*    Bugzilla number:
*    Fixed a bug in setting the mgmt switches
*
*    Revision 1.4  2003/10/30 23:15:15  brian
*    Bugzilla number:
*    Fixed a typo
*
*    Revision 1.3  2003/10/24 21:36:30  brian
*    Bugzilla number:
*    Add access to anafa seeprom
*
*    Revision 1.2  2003/10/24 02:25:04  brian
*    Bugzilla number:
*    Add support for i2c route setting in 270
*
*    Revision 1.1  2003/10/22 00:25:55  brian
*    Bugzilla number:
*    Initial version of sysevent for 270
*
*
***************************************************************************
*                       (c) COPYRIGHT 2000, 2001, 2002
*                           TopSpin Communications, Inc.
*                           Palo Alto, CA
**************************************************************************/
#ifdef RCSID
static const char rcsid[] = "$Id: 270sc_sysevent.c,v 1.51 2005/04/30 03:44:43 brian Exp $";
#endif


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/poll.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/machdep.h>
#include <asm/ocp.h>
#include <linux/delay.h>
#include <i2c-errcode.h>
#include <topspin_types.h>
#include <fpga.h>
#include <system_interrupt_code.h>
#include "270sc_sysevent.h"
#include "ds1672_rtc.h"
#include "dcr.h"
#include "fruerror.h"

extern int global_i2c_debug;
extern int i2c_controller_fast_mode;
extern unsigned char i2c_controller_slave_addr;
extern struct semaphore i2c_lock;


extern int phy_iic_xfer(void *vaddr, unsigned short dev_addr, int access, char *buf, int count, int controller);
extern int phy_iic_init(void *vaddr);

int sysevent_init_module(void);
void sysevent_cleanup_module(void);
int seti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller, unsigned char *mgmt_mux_addr, unsigned char auxselect);
void unseti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller, unsigned char mgmt_mux_addr, unsigned char auxselect);


/* Function Prototypes */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int sysevent_poll(struct file *file, poll_table *wait);
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);


/* global vars */
unsigned char ts270_my_chassis_id;
unsigned char ts270_is_a_node;
volatile int hardware_err_sync_int_active;
volatile int runtime_err_int_active;
volatile int err_fru_id[TS270_MAX_NUM_FRU];
volatile int runtime_err[TS270_MAX_NUM_FRU];
tRUNTIME_FRU_ERROR fruerror[TS270_MAX_NUM_FRU];
unsigned char post_err[TS270_MAX_NUM_FRU][POST_RESULT_LEN];
wait_queue_head_t sysevent_waitq;

extern unsigned char ts270_my_slot_id;
extern int ts270_mgmtmux_override;



/* File Operation Table */
static struct file_operations char_fops = {
    owner: THIS_MODULE,
    read: sysevent_read,
    write: sysevent_write,
    ioctl: sysevent_ioctl,
    poll: sysevent_poll,
};

static wait_queue_head_t masterchange_waitq;
static wait_queue_head_t delay_waitq;
static wait_queue_head_t mgmtmux_waitq;
static wait_queue_head_t optlim_waitq;


static char *ts270_reset_reason;
static volatile int fpga_int_active;
static volatile int mastership_change;
static volatile int smm_fault_int_active;
static volatile int smm_fault_recover_int_active;
static volatile int smm_healthy_int_active;
static volatile int smm_healthy_recover_int_active;
static int specific_mgmt_card;
static int mastership_override;
static int ts270_mgmtmux_override_force; 
static int user_generated_intr_code;
static int syseventinit;



static int timeout = 15;

extern  void *ts270_fpga_base_addr;
static void *i2c0_base_addr;
static void *i2c1_base_addr;
static volatile void *gpio_base_addr;

static void deassert_AB_intr(void)
{
    unsigned char data;

    if (ts270_is_a_node == 0 ) {
        return;
    }

    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    data &= (~0xc0);

    writeb(data, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));

    return;
}


/* assert_AB_intr
 * --------------
 * This function is for a node to call. No ops for a core.
 * A node calls this function to send intr to the cores (both A and B) to notify
 * the cores of a SMM events. The core wil then use ethernet communication to determine what
 * is going on with the SMM on the node.
 */
static void assert_AB_intr(void)
{
    unsigned char data;

    if (ts270_is_a_node == 0 ) {
        return;
    }

    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    data |= 0xc0;

    writeb(data, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    
    return;
}


/* interrupt handler */
static void fpga_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
    unsigned char data;

    /* check for mastership change */
    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
    if ((data & 0x01) != 0) {	/* mastership change */
        /* clear it */
        writeb(0x01, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));        
        mastership_change = 1;
        wake_up_interruptible(&masterchange_waitq);
    }

    /* see if there is still pending intr */
    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET));

    /* check for node err_in intr */
    if ((data & 0x04) != 0) {
        /* go read the err_in change */
        data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG0_OFFSET)); 
        if ((data & 0xf0) != 0) {	/* node err_in intr */
            /* scan to see which node generates err_in intr */
            if ((data & 0x80) != 0) {	/* node 1 */
                err_fru_id[SC_270_SW_NODE1_SLOT_NUM] = SC_270_SW_NODE1_SLOT_NUM;
            }
            if ((data & 0x40) != 0) {   /* node 2 */
                err_fru_id[SC_270_SW_NODE2_SLOT_NUM] = SC_270_SW_NODE2_SLOT_NUM;
            }
            if ((data & 0x20) != 0) {   /* node 3 */
                err_fru_id[SC_270_SW_NODE3_SLOT_NUM] = SC_270_SW_NODE3_SLOT_NUM;
            }
            if ((data & 0x10) != 0) {   /* node 4 */
                err_fru_id[SC_270_SW_NODE4_SLOT_NUM] = SC_270_SW_NODE4_SLOT_NUM;
            }
            /* clear node err_in intr */
            writeb(data & 0xf0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG0_OFFSET));
            hardware_err_sync_int_active = 1; 
            wake_up_interruptible(&sysevent_waitq);
        }

    }

    /* see if there is still pending intr */
    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET));

    /* check for all other intr */
    if ((data & 0x0f) != 0) {
        fpga_int_active = 1;
        wake_up_interruptible(&sysevent_waitq);
    }

}


/* we no longer handle temp faul intr. Polling instead */
static void smm_fault_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
#if 0
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    if (dcr_val & TS270_SMM_FAULT_INT_MASK) {       /* the bit is 1, rising edge - recovery event */
        smm_fault_recover_int_active = 1;
        deassert_AB_intr();
    }
    else {
        smm_fault_int_active = 1;
        assert_AB_intr();
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS270_SMM_FAULT_INT_MASK;
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    wake_up_interruptible(&sysevent_waitq);
#endif

}

/* we no longer support recovery for current and voltage fault. ONce a current or voltage error happens, the card is supposed to be replaced */
static void smm_healthy_int_handler(int this_irq, void *dev_id, struct pt_regs *regs)
{
#if 0
    int dcr_val;

    /* check current polarity */
    dcr_val = mfdcr(UIC0_POLARITY_REG);
    if (dcr_val & TS270_SMM_HEALTHY_INT_MASK) {       /* the bit is 1, rising edge - recovery event */
        /* smm_healthy_recover_int_active = 1; */
        deassert_AB_intr();
    }
    else {
        smm_healthy_int_active = 1;
        assert_AB_intr();
    }

    /* reverse the polarity to catch the other edge */
    dcr_val ^= TS270_SMM_HEALTHY_INT_MASK;
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    wake_up_interruptible(&sysevent_waitq);
#endif

    smm_healthy_int_active = 1;
    wake_up_interruptible(&sysevent_waitq);
    assert_AB_intr();
    deassert_AB_intr();

}


int read_reset_reason(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
    if (offset > 0) {
        return 0;
    }

    len = sprintf(buf, ts270_reset_reason);
    buf[len] = '\n';
    len ++;
    *eof = 1;
    
    return len;
}






int sysevent_init_module(void)
{
    int dcr_val;
    unsigned char data;
    char *sram_sig;
    int fru;
    int i;

    /* init global vars */
    smm_fault_int_active = 0;
    smm_fault_recover_int_active = 0;
    fpga_int_active = 0;
    smm_healthy_int_active = 0;
    smm_healthy_recover_int_active = 0;
    hardware_err_sync_int_active = 0;
    runtime_err_int_active = 0;
    mastership_change = 0;

    specific_mgmt_card = 0;
    mastership_override = 0;
    ts270_mgmtmux_override_force = 0;
    ts270_mgmtmux_override = 0;
    user_generated_intr_code = 0;
    syseventinit = 1;

    for (fru = 0; fru < TS270_MAX_NUM_FRU; fru ++) {
       err_fru_id[fru] =  0;
       runtime_err[fru] =  0;
       for (i = 0; i < POST_RESULT_LEN; i ++) {
           post_err[fru][i] = 0;
       }
       init_fru_error(fru, &fruerror[fru]);
    } 


    /* register char device */
    if (register_chrdev(SYSEVENT_MAJOR, "syseventmodule", &char_fops)) {
        printk(KERN_ERR "Unable to register syseventmodule\n");
        return -EIO;
    }

    create_proc_read_entry("resetreason", 0, NULL, read_reset_reason, NULL);

    sram_sig = (char *)PPC440_SRAM_VIRT_BASE; 
    i2c0_base_addr = ioremap_nocache(I2C_CONTROLLER0_BASE, 0x12); 
    i2c1_base_addr = ioremap_nocache(I2C_CONTROLLER1_BASE, 0x12); 
    gpio_base_addr = (volatile void *)ioremap_nocache(GPIO_BASE_REGADDR, 0x1000);

    /* inspect the sram for reset signature */
    sram_sig[RESET_SIG_LEN] = 0;

    if (!strcmp(sram_sig, SOFTWARE_RESET)) {
        ts270_reset_reason = SOFTWARE_RESET;
    }
    else if (!strcmp(sram_sig, WATCHDOG_RESET)) {
        ts270_reset_reason = WATCHDOG_RESET;
    }
    else {
        ts270_reset_reason = POWER_ON_RESET;
    }

    strcpy(sram_sig, WATCHDOG_RESET);

    /* make sure gpio 11: tri-state disable, no open drain */
        {
            volatile unsigned int data;
            /* disable tri-state - make sure tri-state bit == 1 */
            data = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_TRISTATE_REGADDR)));
            data |= GPIO_11_MASK | GPIO_9_MASK;
            writel(__cpu_to_le32(data), (volatile void *)((unsigned int)gpio_base_addr + GPIO_TRISTATE_REGADDR));

            /* no open drain - make sure open drain bit == 0 */
            data = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OPENDRAIN_REGADDR)));
            data &= ~(GPIO_11_MASK | GPIO_9_MASK);
            writel(__cpu_to_le32(data), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OPENDRAIN_REGADDR));
        }


    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_CHASID_REG_OFFSET));
    ts270_my_chassis_id = (data >> SC_270_FPGA_CHASID_SHIFT) & SC_270_FPGA_CHASID_MASK;
    /* ts270_my_slot_id = data & SC_270_FPGA_FABRIC_SLOT_ID_MASK; */
    if ((ts270_my_slot_id != SC_270_FPGA_CORE1_SLOT_NUM) && (ts270_my_slot_id != SC_270_FPGA_CORE2_SLOT_NUM)) {
        ts270_is_a_node = 1;
    }
    else {
        ts270_is_a_node = 0;
    }



    /* init wait queue */
    init_waitqueue_head(&sysevent_waitq);
    init_waitqueue_head(&masterchange_waitq);
    init_waitqueue_head(&delay_waitq);
    init_waitqueue_head(&mgmtmux_waitq);
    init_waitqueue_head(&optlim_waitq);


    /* configure IRQ4 (Summit Healthy), IRQ5 (FPGA Int), and IRQ6 (Summit Int) to be trigger type, falling edge */
    dcr_val = mfdcr(UIC0_TRIGGER_REG);
    dcr_val |= TS270_UIC0_INT_MASK;
    mtdcr(UIC0_TRIGGER_REG, dcr_val);

    dcr_val = mfdcr(UIC0_POLARITY_REG);
    dcr_val &= (~TS270_UIC0_INT_MASK);
    mtdcr(UIC0_POLARITY_REG, dcr_val);

    /* clear int status */
    dcr_val = mfdcr(UIC0_STATUS_REG);
    dcr_val |= TS270_UIC0_INT_MASK;
    mtdcr(UIC0_STATUS_REG, dcr_val);

    
    /* request interrupt */
    if (request_irq(FPGA_IRQ_NUM, fpga_int_handler, SA_INTERRUPT, "FPGA Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", FPGA_IRQ_NUM);
        return -1;
    }

    if (request_irq(SMM_FAULT_IRQ_NUM, smm_fault_int_handler, SA_INTERRUPT, "Summit Fault Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", SMM_FAULT_IRQ_NUM);
        return -1;
    }

    if (request_irq(SMM_HEALTHY_IRQ_NUM, smm_healthy_int_handler, SA_INTERRUPT, "Summit Healthy Int", NULL)) {
        printk(KERN_ERR "sysevent_init_module: Request IRQ %d Failed\n", SMM_HEALTHY_IRQ_NUM);
        return -1;
    }


    /* init the FPGA. Do this differently for node vs core */
    if (ts270_is_a_node) {
        /* for a node, disable A_pw_fail and B_pw_fail. We do not deal with power fail right now */
        writeb(0x30, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_PWR_INT_CNTL_REG_OFFSET));
        
        /* clear the A_err and B_err change bits here */
        writeb(0x07, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));

        /* disable A_err and B_err */
        writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));

        /* enable master change intr */
        writeb(0x10, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET)); 
    }
    else {	/* a core, same for both master and non-master */
        
        /* Presence change bit */
        writeb(0x01, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG2_OFFSET));

        /* enable presence and master change intr */
        writeb(0x90, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET)); 

        /* Power supplies: clear Change and disable intr */
        writeb(0xf0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_PWR_INT_CNTL_REG_OFFSET));

        /* Module and LIM ERRn */
        writeb(0xff, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG0_OFFSET));
        writeb(0xff, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_ERR_EDGEDET_REG1_OFFSET));

        /* disable Error Intr from  all*/
        writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_ERR_INT_EN_REG0_OFFSET));
        writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_ERR_INT_EN_REG1_OFFSET));

        /* Module and LIM INTn */
        writeb(0xff, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_INT_EDGEDET_REG0_OFFSET));
        writeb(0xff, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_INT_EDGEDET_REG1_OFFSET));

        /* disable Intr from all cards */
        writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_INT_CPUINT_EN_REG0_OFFSET));
        writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_INT_CPUINT_EN_REG1_OFFSET));

        /* C_ERR */
        writeb(0x07, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));

        /* disable C_err */
        writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
    }

    return 0;
}


void sysevent_cleanup_module(void)
{

    unregister_chrdev(SYSEVENT_MAJOR, "syseventmodule");
    remove_proc_entry("resetreason", NULL);
    iounmap(i2c0_base_addr);
    iounmap(i2c1_base_addr);
    iounmap((void *)gpio_base_addr);
    free_irq(FPGA_IRQ_NUM, NULL);
    free_irq(SMM_FAULT_IRQ_NUM, NULL);
    free_irq(SMM_HEALTHY_IRQ_NUM, NULL);
}


#define TS270_POLLING_NUM      4
const int polling_code[TS270_POLLING_NUM] = {EXT_INT_CODE, TS270_FPGASYNC_INT_CODE, TS270_FRUDATASYNC_INT_CODE, SLAVE_CC_INT_CODE};


/* Read routine */
static ssize_t sysevent_read(struct file *file, char *buf, size_t count, loff_t *offset)
{
    int interrupt_code;
    int delay;
    int minor;
    struct inode *inode = file->f_dentry->d_inode;
    unsigned char data;
    static int i = 0;
    int fru;
    int len;
    int runtime_intcode;

    minor = MINOR(inode->i_rdev);

    if (minor == SYSEVENT_MASTERCHANGE_MINOR) {
        while (1) {
            if (mastership_change == 0) {
                interruptible_sleep_on_timeout(&masterchange_waitq, HZ);
            }
            else {
                break;
            }

            data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
            if ((data & 0x01) != 0) {   /* mastership change */
                /* clear it */
                writeb(0x01, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
                mastership_change = 1;
            }
        }
        mastership_change = 0;
        interrupt_code = MASTERCHANGE_INT_CODE;
        copy_to_user(buf, &interrupt_code, sizeof(int));
        return sizeof(int);
    }
    else if (minor != 0) {
        if (count > POST_RESULT_LEN) {
            len = POST_RESULT_LEN;
        }
        else {
            len = count;
        }
        copy_to_user(buf, &post_err[minor][0], len);

        return len;
    }
 
    
    /* if here, default minor == 0 */ 

    if ((smm_fault_int_active == 0) && 
        (smm_fault_recover_int_active == 0) && 
        (fpga_int_active == 0) && 
        (smm_healthy_int_active == 0) && 
        (hardware_err_sync_int_active == 0) && 
        (runtime_err_int_active == 0) && 
        (smm_healthy_recover_int_active == 0)) {
        interruptible_sleep_on_timeout(&sysevent_waitq, timeout * HZ);
    }

    if (hardware_err_sync_int_active == 1) {
        interrupt_code = HARDWARE_ERR_SYNC_INT_CODE;
        /* scan to see which fru has errors */
        for (fru = 1; fru < TS270_MAX_NUM_FRU; fru ++) {
            if (err_fru_id[fru] != 0) {
                interrupt_code |= err_fru_id[fru] << SYSTEM_INT_SHIFT;
                err_fru_id[fru] =  0;
                hardware_err_sync_int_active = 0;
                break;
            }
        }  
        /* scan one more time to check any other pending errors */
        for (fru = 1; fru < TS270_MAX_NUM_FRU; fru ++) {
            if (err_fru_id[fru] != 0) {
                hardware_err_sync_int_active = 1;
                break;
            }
        }  

        delay = 0;
        goto sysevent_read_return;
    }

    if (runtime_err_int_active == 1) {
        interrupt_code = HARDWARE_ERR_SYNC_INT_CODE;
        for (fru = 1; fru < TS270_MAX_NUM_FRU; fru ++) {
            if (runtime_err[fru] != 0) {
                runtime_intcode = FRU_ERR_CODE | 0x80 | fru;
                interrupt_code |= runtime_intcode << SYSTEM_INT_SHIFT;
                runtime_err[fru] = 0;
                runtime_err_int_active = 0;
                break;
            }
        }
        /* scan one more time to check any other pending errors */
        for (fru = 1; fru < TS270_MAX_NUM_FRU; fru ++) {
            if (runtime_err[fru] != 0) {
                runtime_err_int_active = 1;
                break;
            }
        }

        delay = 0;
        goto sysevent_read_return;
    }



    if (smm_fault_int_active == 1) {
        interrupt_code = SMM_FAULT_INT_CODE;
        smm_fault_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_fault_recover_int_active == 1) {
        interrupt_code = SMM_FAULT_RECOVER_INT_CODE;
        smm_fault_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }


    if (fpga_int_active == 1) {
        interrupt_code = FPGA_INT_CODE;
        fpga_int_active = 0;
        delay = 1;
        goto sysevent_read_return;
    }

    if (smm_healthy_int_active == 1) {
        interrupt_code = SMM_HEALTHY_INT_CODE;
        smm_healthy_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (smm_healthy_recover_int_active == 1) {
        interrupt_code = SMM_HEALTHY_RECOVER_INT_CODE;
        smm_healthy_recover_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
    if ((data & 0x01) != 0) {   /* mastership change */
        /* clear it */
        writeb(0x01, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG0_OFFSET));
        mastership_change = 1;
    }

    data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MISC_INT_CNTL_REG_OFFSET));
    if ((data & 0x0f) != 0) {
        interrupt_code = FPGA_INT_CODE;
        fpga_int_active = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    if (user_generated_intr_code != 0) {
        interrupt_code = user_generated_intr_code;
        user_generated_intr_code = 0;
        delay = 0;
        goto sysevent_read_return;
    }

    delay = 0;

    /* return with interrupt code == polling code */
    interrupt_code = polling_code[i];

    i ++;
    if (i == TS270_POLLING_NUM) {
        i = 0;
    }



sysevent_read_return:

    /* delay before returning */
    interruptible_sleep_on_timeout(&delay_waitq, delay * HZ);
    copy_to_user(buf, &interrupt_code, sizeof(int));
    return sizeof(int);

}


/* Write routine
 * -------------
 * Users want to clear a bit mask in the interrupt code. The value should have the bit zero, all other bits are 1
 */
static ssize_t sysevent_write(struct file *file, const char *buf, size_t count, loff_t *offset)
{
    int value;
    int controller;
    unsigned char fruid;
    int minor;
    struct inode *inode = file->f_dentry->d_inode;
    int len;

    minor = MINOR(inode->i_rdev);

    if (minor != 0) {
        if (count > POST_RESULT_LEN) {
            len = POST_RESULT_LEN;
        }
        else {
            len = count;
        }
        copy_from_user(&post_err[minor][0], buf, len);

        return len;
    }

    /* if here default minor == 0 */

    copy_from_user(&value, buf, sizeof(int));

    if ((value & I2C_DEBUG_CODE) == I2C_DEBUG_CODE) {
        global_i2c_debug = value & 0xff;
    }

    if ((value & I2C_CLOCK_CODE) == I2C_CLOCK_CODE) {
        i2c_controller_fast_mode = (value >> SYSEVENTWRITE_I2CCLOCK_MODE_SHIFT) & SYSEVENTWRITE_I2CCLOCK_MODE_MASK;
        i2c_controller_slave_addr = (value >> SYSEVENTWRITE_I2CCLOCK_SLAVE_SHIFT) & SYSEVENTWRITE_I2CCLOCK_SLAVE_MASK;
        controller = (value >> SYSEVENTWRITE_I2CCLOCK_BUS_SHIFT) & SYSEVENTWRITE_I2CCLOCK_BUS_MASK;
        down(&i2c_lock);
        if (controller == 0) {
            phy_iic_init(i2c0_base_addr);
        }
        else {
            phy_iic_init(i2c1_base_addr);
        } 
        up(&i2c_lock);
    }

    if ((value & FRU_ERR_CODE) == FRU_ERR_CODE) {
        fruid = value & SYSEVENTWRITE_FRUID_MASK;
        err_fru_id[fruid] = value & 0xffff;
        hardware_err_sync_int_active = 1;
        wake_up_interruptible(&sysevent_waitq);
    }

    if ((value & TS270_SPECIFY_I2CROUTE) == TS270_SPECIFY_I2CROUTE) {
        specific_mgmt_card = value & 0xff;
    }

    if ((value & TS270_MASTER_OVERRIDE) == TS270_MASTER_OVERRIDE) {
        mastership_override = value & 0xff;
    }

    if ((value & TS270_MGMTMUX_OVERRIDE) == TS270_MGMTMUX_OVERRIDE) {
        if ((value & 0xff) == 0) {
            ts270_mgmtmux_override_force = 0;
            ts270_mgmtmux_override = 0;
        }
        else {
            ts270_mgmtmux_override_force = 1;
            ts270_mgmtmux_override = 1; 
        } 
    }

    if ((value & TS270_GENERATE_EVENT) == TS270_GENERATE_EVENT) {
        user_generated_intr_code = value & 0xff;
        wake_up_interruptible(&sysevent_waitq);
    }


    return sizeof(int);
}

/* Ioctl routine */
static int sysevent_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    int command;
    int fru_location;
    tRUNTIME_ERR_DATA entrydata;
    unsigned short errcode_key;

    fru_location = cmd & SYSEVENTWRITE_FRUID_MASK;
    command = cmd >> 8;

    switch(command) {
        case TS_FRUERR_ADD:
            if (copy_from_user(&entrydata, (tRUNTIME_ERR_DATA *)arg, sizeof(tRUNTIME_ERR_DATA))) {
                return -EFAULT;
            }
            add_entry_to_fru_error(fru_location, entrydata, &fruerror[fru_location]);
            runtime_err[fru_location] = 1;
            runtime_err_int_active = 1;
            wake_up_interruptible(&sysevent_waitq);
            break;
        case TS_FRUERR_ADD_PROCESSED:
            if (copy_from_user(&entrydata, (tRUNTIME_ERR_DATA *)arg, sizeof(tRUNTIME_ERR_DATA))) {
                return -EFAULT;
            }
            add_entry_to_fru_error_processed(fru_location, entrydata, &fruerror[fru_location]);
            break;
        case TS_FRUERR_REMOVE:
            if (copy_from_user(&errcode_key, (unsigned short *)arg, sizeof(unsigned short))) {
                return -EFAULT;
            }
            remove_entry_from_fru_error(fru_location, errcode_key, &fruerror[fru_location]);
            break;
        case TS_FRUERR_GET_ALL:
            {
                unsigned char buf[sizeof(int) + sizeof(tRUNTIME_ERR_DATA) * TS_RUNTIME_ERR_MAX_ENTRY];
                tRUNTIME_ERR_DATA *entrydata;
                int *returned_num_entries;

                returned_num_entries = (int *)&buf[0];
                entrydata = (tRUNTIME_ERR_DATA *)&buf[sizeof(int)];

                retrieve_all_processed_entry(fru_location, TS_RUNTIME_ERR_MAX_ENTRY, returned_num_entries, entrydata, &fruerror[fru_location]);
                copy_to_user((unsigned char *)arg, buf, sizeof(int) + sizeof(tRUNTIME_ERR_DATA) * (*returned_num_entries));
                return 0;
                break;
            }
        case TS_FRUERR_RESET:
            reset_fru_error(fru_location, &fruerror[fru_location]);
            break;
        case TS_FRUERR_PROCESS:
            {
                tRUNTIME_ERR_DATA entrydata;

                entrydata.errcode_key = 0;	/* when users get back the entry, this field should be non-zero */

                process_entry_from_fru_error(fru_location, &entrydata, &fruerror[fru_location]);
                copy_to_user((tRUNTIME_ERR_DATA *)arg, &entrydata, sizeof(tRUNTIME_ERR_DATA));
                return 0;
                break;
            }
        case TS_FRUERR_SET_ALL:
            {
                unsigned char buf[sizeof(int) + sizeof(tRUNTIME_ERR_DATA) * TS_RUNTIME_ERR_MAX_ENTRY];
                tRUNTIME_ERR_DATA *entrydata;
                int num_entries;

                if (copy_from_user(buf, (unsigned char *)arg, sizeof(int) + sizeof(tRUNTIME_ERR_DATA) * TS_RUNTIME_ERR_MAX_ENTRY)) {
                    return -EFAULT;
                }
                num_entries = *((int *)&buf[0]);
                entrydata = (tRUNTIME_ERR_DATA *)&buf[sizeof(int)];

                make_fru_error(fru_location, num_entries, entrydata, &fruerror[fru_location]);
                return 0;
                break;
            }
        default:
            break;
    }

    return 0;
}


static unsigned int sysevent_poll(struct file *file, poll_table *wait)
{
    return 0;
}





int read_mgmt_mux(void *vaddr, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller, unsigned char mgmt_mux_addr, unsigned char *readdata)
{
    unsigned char data;
    int ret;

    /* read the mux to see the control bits */
    data = MGMT_MUX_CONTROL_REG_OFFSET;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, READACCESS, readdata, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    return 0;
}






void mgmt_mux_check(unsigned char data, int *is_channel_off, int *do_i_have_it, unsigned char *writedata)
{
    unsigned char nmygrab; 
    unsigned char nmyon; 
    unsigned char onoff_status; 

    onoff_status = data & 0x0c;

    if ((onoff_status == 0x0c) || (onoff_status == 0x0)) {	/* channel is off */
        /* if we are in override mode, try to turn it off because the other master appears to release the channel now */
        if ((syseventinit == 1) && (ts270_mgmtmux_override > 0) && (ts270_mgmtmux_override_force == 0)) {
            ts270_mgmtmux_override = 0;
            if (global_i2c_debug >= 1) {
                printk(KERN_DEBUG "Turn off mux override mode\n");
            }
        }
    } 


  if (ts270_mgmtmux_override > 0) {

    *do_i_have_it = 1;
    *is_channel_off = 1;
    *writedata = data;

    /* make sure the channel is on */
    if ((data & 0x0c) == 0x0c) {
        *do_i_have_it = 0;
        *writedata &= (~0x04);
    }

    if ((data & 0x0c) == 0) {
        *do_i_have_it = 0;
        *writedata |= 0x04;
    }

    /* make sure I can grab the channel */
    if ((data & 0x03) == 0x01) {
        *do_i_have_it = 0;
        *writedata &= (~0x01);
    }

    if ((data & 0x03) == 0x02) {
        *do_i_have_it = 0;
        *writedata |= 0x01;
    }

    return;
  }








    nmygrab = data & 0x02;
    nmyon = data & 0x08;

    if ((data & 0x0c) == 0x0c) {
        *is_channel_off = 1;
        *do_i_have_it = 0;
        /* writedata for backoff */
        *writedata = 0;
        return;
    }

    if ((data & 0x0c) == 0x0) {
        *is_channel_off = 1;
        *do_i_have_it = 0;
        /* writedata in case of force stealing the channel */
        if (nmygrab == 0) {
            *writedata = 0x4;
        }
        else {
            *writedata = 0x5;
        }
        return;
    }

    /* if here, the channel is on */
    if ((ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) && ((data & 0x03) == 0x0)) {
        *is_channel_off = 0;
        *do_i_have_it = 1;
        return;
    }

    if ((ts270_my_slot_id == SC_270_FPGA_CORE2_SLOT_NUM) && ((data & 0x03) == 0x3)) {
        *is_channel_off = 0;
        *do_i_have_it = 1;
        return;
    }
 

    *is_channel_off = 0;
    *do_i_have_it = 0;
    /* writedata in case of force stealing the channel */
    if (nmygrab == 0) {
        if (nmyon == 0) {
            *writedata = 0x4;
        }
        else {
            *writedata = 0x0;
        }
    }
    else {
        if (nmyon == 0) {
            *writedata = 0x5;
        }
        else {
            *writedata = 0x1;
        }
    }

    return;
}


 



int compete_for_mgmt_mux(void *vaddr, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller, unsigned char mgmt_mux_addr)
{
    int ret;
    unsigned char data;
    unsigned char writedata;
    unsigned char buf[4];
    int is_channel_off;
    int do_i_have_it;
    int count = 0;

  while (1) {

    if ((ret = read_mgmt_mux(vaddr, iic_xfer, controller, mgmt_mux_addr, &data)) != 0) {
        return ret;
    }


    /* inspect the read data */
    mgmt_mux_check(data, &is_channel_off, &do_i_have_it, &writedata);

    if (do_i_have_it == 1) {
        return 0;		/* no need to do anything else */
    }


    /* if here, either the channel is off or the other master has it. Check the later case first */
    while (is_channel_off == 0) {	/* channel is on, the other master is using it */
        if (syseventinit == 1) {
            interruptible_sleep_on_timeout(&mgmtmux_waitq, 1);
        }
        else {
            udelay(10000);
        } 
        count ++;
        if (count > 50) {	/* other other channel owns it for too long (max 500ms) , give up */
            if (global_i2c_debug >= 1) {
                printk(KERN_DEBUG "Timneout waiting for mgmt mux %x to be off, data = %x\n", mgmt_mux_addr, data);
                printk(KERN_DEBUG "Turn on mux override mode\n");
            }
            ts270_mgmtmux_override = 1;
        }
        if ((ret = read_mgmt_mux(vaddr, iic_xfer, controller, mgmt_mux_addr, &data)) != 0) {
            return ret;
        }
        mgmt_mux_check(data, &is_channel_off, &do_i_have_it, &writedata);

    }	/* end while */



    /* if here, the channel is off or timeout. Grab the channel */
    buf[0] = MGMT_MUX_CONTROL_REG_OFFSET;
    buf[1] = writedata;

    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, WRITEACCESS, buf, 2, controller)) != 2) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    if (ts270_mgmtmux_override == 0) {		/* no need to wait if override */
        if (syseventinit == 1) {
            interruptible_sleep_on_timeout(&mgmtmux_waitq, HZ/100);
        }
        else {
            udelay(10000);
        }
    }
    else {
        break;
    }

    /* come back to the loop to check if I really get the channel */

 }	/* while loop */

    return 0;

}





/* unseti2croute
 * --------------
 * disconnect the mgmt mux to allow 2 cores to compete for the same mgmt mux.
 */
void unseti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller, unsigned char mgmt_mux_addr, unsigned char auxselect)
{
    unsigned char buf[4];
    unsigned char lim_num;
#if 0
    unsigned char data;
    int ret;
#endif


    if (controller == 1) {      /* on 270, controller 1 is NOT connected to the i2c switch */
        return;
    }


    /* to avoid recursive calls */
    if (mgmt_mux_addr == 0) {
        return;
    }

#if 0

    if ((ret = read_mgmt_mux(vaddr, iic_xfer, controller, mgmt_mux_addr, &data)) != 0) {
        return;
    }

    buf[0] = MGMT_MUX_CONTROL_REG_OFFSET;

    if ((data & 0x08) == 0) {
        buf[1] = 0;	 
    }
    else {
        buf[1] = 4;                
    }

#endif

    /* if we have arbitrated with the PLD on optical LIMs, we need to release the request now */
    auxselect &= 0x0f;
    if ((auxselect >= 1) && (auxselect <= 8)) {
        lim_num = (muxselect >> 2) & 0x03;      /* bits 3:2 */ 
        /* de-assert access request */
        buf[0] = OPT_LIM_IO_OUTPORT0_OFFSET;
        buf[1] = 0xff; 
        if (lim_num == 1) {
            iic_xfer(vaddr, (0xff << 8) | OPT_LIM1_OUT_I2C_ADDR, WRITEACCESS, buf, 2, controller);
        }
        else {
            iic_xfer(vaddr, (0xff << 8) | OPT_LIM2_OUT_I2C_ADDR, WRITEACCESS, buf, 2, controller);
        }
    }

    buf[0] = MGMT_MUX_CONTROL_REG_OFFSET;
    buf[1] = 0;

    iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, WRITEACCESS, buf, 2, controller);


    return;
}





/* deifne the mux select value to set i2c route
 * This mux select is a one-byte value:
 * - bit 0: 1 for on-board, 0 for off-board
 * - bit 1: select mgmt card. 0 for mgmt card 1, 1 for mgmt card 2.
 * If the mgmt switch value is NOT to select on-board mgmt card devices, this bit is used to determine which mgmt card to go thru.
 * 0 means default route, 1 means alternate route.
 * Default route is: core 1 goes thru mgmt card 1, core 2 goes thru mgmt card 2.
 * Alternate route is the opposite of the default route. Alternate route is used by upper level application for error handling: when the default route fails.
 * Another way to control which mgmt card to go thru is to use syseventw, code TS270_SPECIFY_I2CROUTE (defined in all/common/include/system_interrupt_code.h ). 
 * - bit [3:2]: lim number - 0 for node card, 1 for min 1, 2 for lim 2
 * - bit [7:4]: management switch value.
 * auxselect was added to support switch on optical LIMs.
 */

int seti2croute(void *vaddr, unsigned char muxselect, int (*iic_xfer)(void *, unsigned short, int, char *, int, int), int controller, unsigned char *changed_mgmt_mux_addr, unsigned char auxselect)
{
    unsigned char presence;
    unsigned char data;
#ifdef TEST_INTR
    unsigned char errout;
#endif
    volatile unsigned char intr_in;
    unsigned char mgmt_mux_addr;
    unsigned char card_mux_addr;
    unsigned char mgmt_switch_addr;
    unsigned char other_mgmt_switch_addr;
    int need_to_change_mux;
    int need_to_set_node_mux = 0;
    unsigned char mgmtswitch;
    unsigned char lim_num;
    int ret;
    int count;
    unsigned char buf[4];

    *changed_mgmt_mux_addr = 0;


#if 0
    if (syseventinit == 1) {
        writeb(0x1f, (void *)((int)i2c0_base_addr + I2C_CONTROLLER_MODE_CONTROL));
        i2c_controller_fast_mode = 1;
    }
#endif

    if (muxselect == 0xff) {
        return 0;
    }

    if (controller == 1) {		/* on 270, controller 1 is NOT connected to the i2c switch */
        return 0;
    }

    mgmtswitch = (muxselect >> 4) & 0x0f;   /* bits 7:4 */

    if ((muxselect & 0x1) == 0x1) {     /* bit 0 == 1 for local on-board devices */
        if (mgmtswitch == 14) {			/* anafa2 seeprom 1 */
            data = 8;					/* channel 3 */
        }
        else if (mgmtswitch == 15) {	/* anafa2 seeprom 2 */
            data = 16;					/* channel 4 */
        }
        else {							/* summit, etc. */
            data = 4;					/* channel 2 */
#if 0
            writeb(0x0f, (void *)((int)i2c0_base_addr + I2C_CONTROLLER_MODE_CONTROL));
            i2c_controller_fast_mode = 0;
#endif
        }
        if ((ret = iic_xfer(vaddr, (0xff << 8) | LOCAL_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
            return IIC_ERR_LOCALSWITCH_ERR;
        }
        else {
            return 0;
        }
    }

    /* if here, we are going to/thru a mgmt card. Only core cards can go thru mgmt card */
    if (ts270_is_a_node) {
        return 0;
    }

    presence = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_PRES_REG2_OFFSET));
    if ((presence & 0x0c) == 0) {		/* no mgmt cards are present */
        return IIC_ERR_MGMTS_MISSING;
    }


  if (mgmtswitch == 6) {		/* we are targeting devices on mgmt card. In this case, use bit 1 to determine which card */

    if ((muxselect & 0x2) == 0x2) {     /* bit 1 = 1 for mgmt card 1 */
        mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
        data = 2;	/* channel 1 */
    }
    else {								/* bit 1 = 0 for mgmt card 0 */
        mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
        data = 1;	/* channel 0 */
    }

  }
  else {	/* we are using mgmt card as a mean to reach node/lim cards. In this case, use bit 1 to determine ewither default route or alternate route */

    if ((presence & 0x0c) == 0x0c) {	/* both mgmt cards are present */
        if ((muxselect & 0x2) == 0x2) {     /* bit 1 = 1 for alternate route */
            if (ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) {
                mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
                data = 2;   /* channel 1 */
            }
            else {
                mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
                data = 1;   /* channel 0 */
            } 
        }
        else {								/* bit 1 = 0 for default route */
            if (ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) {
                mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
                data = 1;   /* channel 0 */
            }
            else {
                mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
                data = 2;   /* channel 1 */
            }
        }
    }

    if ((presence & 0x0c) == 0x08) {	/* only mgmtA present */
        mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
        data = 1;   /* channel 0 */
    }

    if ((presence & 0x0c) == 0x04) {    /* only mgmtB present */
        mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
        data = 2;   /* channel 1 */
    }


    /* if specific route is set using syseventw, it will take precedence */
    if (specific_mgmt_card == 15) {
        mgmt_mux_addr = MGMT0_MUX_I2C_ADDR;
        data = 1;   /* channel 0 */
    }

    if (specific_mgmt_card == 16) {
        mgmt_mux_addr = MGMT1_MUX_I2C_ADDR;
        data = 2;   /* channel 1 */
    }

  }		/* end if of using mgmt cards to reach other devices */

    /* double check mgmt card presence again */
    if ((mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) && ((presence & 0x08) != 0x08)) {		/* select mgmt A, but mgmt A not present */
        return IIC_ERR_MGMT0_MISSING;
    }

    if ((mgmt_mux_addr == MGMT1_MUX_I2C_ADDR) && ((presence & 0x04) != 0x04)) {		/* select mgmt B, but mgmt B not present */
        return IIC_ERR_MGMT1_MISSING;
    }
         

    /* if I want to access the other mgmt card, check if the other is asserting error out, which indicates a PICInit in progress */
    if (((ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) && (mgmt_mux_addr == MGMT1_MUX_I2C_ADDR)) ||
        ((ts270_my_slot_id == SC_270_FPGA_CORE2_SLOT_NUM) && (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR)) ) {
#ifdef TEST_INTR
        errout = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET)); 
        if ((errout & 0x01) != 0) {   /* error active */
            return IIC_ERR_PIC_UPDATE;
        }
#else
        intr_in = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
        if ((intr_in & 0x40) != 0) {   /* error active */
            return IIC_ERR_PIC_UPDATE;
        }
#endif
    }

    if (ts270_mgmtmux_override_force > 0) {
        intr_in = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
        if ((intr_in & 0x40) != 0) {	/* the other core is requesting access */
            /* yield to the other core */
#ifdef TEST_INTR
            writeb(0x07, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET)); 
#else
            writeb(0x0, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET)); 
#endif
            /* wait until the other removes the request */
            count = 0;
            while ((intr_in & 0x40) != 0) {   
                interruptible_sleep_on_timeout(&mgmtmux_waitq, 1);
                count ++;
                if (count > 50) {   /* other other channel owns it for too long (max 500ms) , give up */
                    break; 
                }
                intr_in = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
            }
#ifdef TEST_INTR
            writeb(0x40, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG1_OFFSET));
#else
            writeb(0x80, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_SYSERR_INTPIN_REG_OFFSET));
#endif
        }
    }

 
            
            
        


    if ((ret = iic_xfer(vaddr, (0xff << 8) | LOCAL_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_LOCALSWITCH_ERR;
    }


    *changed_mgmt_mux_addr = mgmt_mux_addr;
    ret = compete_for_mgmt_mux(vaddr, iic_xfer, controller, mgmt_mux_addr);
    if (ret != 0) {
        return ret;
    }



    /* now set the switch on the mgmt card */
    if (mgmtswitch < 4) {   /* try to talk to node cards */
        mgmt_switch_addr = MGMT_SWITCH0_I2C_ADDR;
        other_mgmt_switch_addr = MGMT_SWITCH1_I2C_ADDR;
        need_to_set_node_mux = 1;
    }
    else {
        mgmt_switch_addr = MGMT_SWITCH1_I2C_ADDR;
        other_mgmt_switch_addr = MGMT_SWITCH0_I2C_ADDR;
        mgmtswitch -= 4;
        /* no mux for on-board mgmt card (channel 2) and for power supplies (channel 5 & 6) */
        if ((mgmtswitch == 2) || (mgmtswitch == 5) || (mgmtswitch == 6)) {
            need_to_set_node_mux = 0;
        }
        else {
            need_to_set_node_mux = 1;
        }
    }

    /* deselect the other switch */
    data = 0;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | other_mgmt_switch_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    /* then select this switch */
    data = 1 << mgmtswitch;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_switch_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        if (mgmt_mux_addr == MGMT0_MUX_I2C_ADDR) {
            return IIC_ERR_MGMT0_ERR;
        }
        else {
            return IIC_ERR_MGMT1_ERR;
        }
    }

    /* if not trying to talk to node cards or the core cards, we are done */
    if (need_to_set_node_mux == 0) {
        return 0;
    }

    /* if here, we are trying to get to the node or the core cards */
    lim_num = (muxselect >> 2) & 0x03;		/* bits 3:2 */
    if (lim_num == 0) {
        card_mux_addr = NODE_MUX_I2C_ADDR;
    }
    else if (lim_num == 1) {
        card_mux_addr = LIM1_MUX_I2C_ADDR;
    }
    else {
        card_mux_addr = LIM2_MUX_I2C_ADDR;
    }


    /* now read the mux to see the control bits  */
    data = MGMT_MUX_CONTROL_REG_OFFSET;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | card_mux_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_TARGETCARD_MUX_ERR;
    }

    if ((ret = iic_xfer(vaddr, (0xff << 8) | card_mux_addr, READACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_TARGETCARD_MUX_ERR;
    }

    need_to_change_mux = 0;
    /* here is the format of the control: bit 0 = mygrab, bit 1 = Nmygrab, bit 2 = myonoff, bit 3 = Nmyonoff */
    /* make sure the channel is on */
    if ((data & 0x0c) == 0x0c) {
        need_to_change_mux = 1;
        data &= (~0x04);
    }

    if ((data & 0x0c) == 0) {
        need_to_change_mux = 1;
        data |= 0x04;
    }

    /* make sure I can grab the channel */
    if ((data & 0x03) == 0x01) {
        need_to_change_mux = 1;
        data &= (~0x01);
    }

    if ((data & 0x03) == 0x02) {
        need_to_change_mux = 1;
        data |= 0x01;
    }

    if (need_to_change_mux) {
        buf[0] = MGMT_MUX_CONTROL_REG_OFFSET;
        buf[1] = data;
        if ((ret = iic_xfer(vaddr, (0xff << 8) | card_mux_addr, WRITEACCESS, buf, 2, controller)) != 2) {
            return IIC_ERR_TARGETCARD_MUX_ERR;
        }
    }

    /* final switch for LIMs */
    if (lim_num == 0) {
        return 0;
    }

#if 0
    data = 1;       /* all LIM on-board devices reside on channel 0 for now */
#endif

    /* auxselect is used to set the switch on optical LIMs.
     * For now, let us use the lower 4 bits of auxselect to access optical LIMs.
     * Optical LIMs have devices on channel 1, some of which are in front of the arbitration mux. 
     * Others (8 module i2c devices) are behind the arbitration mux.
     */

    auxselect &= 0x0f;	/* ignore upper bits for now */

    if (auxselect == 0) {
        data = 1;	/* channel 0, for older devices */
    }
    else {
        data = 2;	/* channel 1, for new devices */
    }

    if (lim_num == 1) {
        if ((ret = iic_xfer(vaddr, (0xff << 8) | LIM1_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
            return IIC_ERR_LIM_SWITCH_ERR;
        }
    }
    else {
        if ((ret = iic_xfer(vaddr, (0xff << 8) | LIM2_SWITCH_I2C_ADDR, WRITEACCESS, &data, 1, controller)) != 1) {
            return IIC_ERR_LIM_SWITCH_ERR;
        }
    }

    /* we assign auxselect from 1 to 8 to select 8 optical modules behind the arbitration mux.
     * If not the optical modules, return now.
     */
    if ((auxselect == 0) || (auxselect > 8)) {
        return 0;
    }

    /* if here, we need to arbitrate with the PLD to get accesses to the optical modules */
  {
    unsigned char io_expander_addr;
    unsigned char second_lim_switch;

    if (lim_num == 1) {
        io_expander_addr = OPT_LIM1_OUT_I2C_ADDR;
        second_lim_switch = OPT_LIM1_SECOND_SWITCH_I2C_ADDR;
    }
    else {
        io_expander_addr = OPT_LIM2_OUT_I2C_ADDR;
        second_lim_switch = OPT_LIM2_SECOND_SWITCH_I2C_ADDR;
    }
    
    
    /* first make sure IO 0 is set for output */
    buf[0] = OPT_LIM_IO_CONFIG0_OFFSET;
    buf[1] = 0xfe;	/* bit 0 as output */
    if ((ret = iic_xfer(vaddr, (0xff << 8) | io_expander_addr, WRITEACCESS, buf, 2, controller)) != 2) {
        return IIC_ERR_OPT_LIM_IO_ERR;
    }

    /* assert access request */
    buf[0] = OPT_LIM_IO_OUTPORT0_OFFSET;
    buf[1] = 0xfe;  /* low at bit 0 */
    if ((ret = iic_xfer(vaddr, (0xff << 8) | io_expander_addr, WRITEACCESS, buf, 2, controller)) != 2) {
        return IIC_ERR_OPT_LIM_IO_ERR;
    }

    /* sleep for 10 ms */
    interruptible_sleep_on_timeout(&optlim_waitq, 1);
    
    /* check if granted */
    data = OPT_LIM_IO_INPORT0_OFFSET;
    if ((ret = iic_xfer(vaddr, (0xff << 8) | io_expander_addr, WRITEACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_OPT_LIM_IO_ERR;
    }

    if ((ret = iic_xfer(vaddr, (0xff << 8) | mgmt_mux_addr, READACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_OPT_LIM_IO_ERR;
    }

    /* grant is on bit 1, active low */
    if ((data & 0x02) != 0) {	/* not granted */
        return IIC_ERR_OPT_LIM_ARBITRATE_ERR;		/* we will remove the request in unseti2croute */
    }

    /* now we get the grant, go set the second LIM switch with auxselect value from 1 to 8 */
    data = 1 << (auxselect - 1);
    if ((ret = iic_xfer(vaddr, (0xff << 8) | second_lim_switch, WRITEACCESS, &data, 1, controller)) != 1) {
        return IIC_ERR_OPT_LIM_SECOND_SWITCH_ERR;
    }
  }


    return 0;
}


/* init_ts_watchdog
 * ------------
 * Start the hardware watchdog.
 */
void init_ts_watchdog(void)
{
    volatile unsigned int gpio_output;

    gpio_output = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR)));

    /* write 1 */
    gpio_output |= GPIO_11_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    /* write 0 */
    gpio_output &= ~GPIO_11_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    return;
}


/* ts270_anafa_reset
 * ------------
 */
void ts270_anafa_reset(void)
{
    volatile unsigned int gpio_output;

    gpio_output = __le32_to_cpu(readl((volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR)));

    /* write 0 */
    gpio_output &= ~GPIO_9_MASK;
    gpio_output &= ~GPIO_10_MASK;
    writel(__cpu_to_le32(gpio_output), (volatile void *)((unsigned int)gpio_base_addr + GPIO_OUTPUT_REGADDR));

    return;
}


/* ts270_disable_dma
 * -----------------
 */
void ts270_disable_dma(void)
{
    writeb(0x3, (void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_DMA_CONTROL_OFFSET));
}


int is_i2c_access_allowed(unsigned char muxselect) 
{
    unsigned char data;

    /* if i am not the master, restrict access to system i2c resources */
    if ((muxselect == TS270_FPGA_MUX_SELECT_SLOT1) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT2) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT3) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT4) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT5) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT6) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT7) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT8) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT9) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT10) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT13) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT14) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT15) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT16) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT17) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT18) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT19) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT20) ||
        (muxselect == TS270_FPGA_MUX_SELECT_SLOT21)) {

        if (ts270_is_a_node == 1) {     /* this is a node */
            return IIC_ERR_NON_MASTER;
        }

        /* if here, i am a core. Check for mastership */
        if (mastership_override > 0) {
            return 0;
        }

        data = readb((void *) ((int) ts270_fpga_base_addr + SC_270_FPGA_MOD_CHASID_REG_OFFSET));
        if ((data & 0x80) == 0) {   /* not the master */
            /* if i am not the master, I can only be allowed to access my own mgmt card */
            if ((ts270_my_slot_id == SC_270_FPGA_CORE1_SLOT_NUM) && (muxselect == TS270_FPGA_MUX_SELECT_SLOT15)) {
                return 0;
            }
            if ((ts270_my_slot_id == SC_270_FPGA_CORE2_SLOT_NUM) && (muxselect == TS270_FPGA_MUX_SELECT_SLOT16)) {
                return 0;
            }
            return IIC_ERR_NON_MASTER;
        }
    }

    return 0;
}





