/*
 * arch/ppc/platforms/ebony.c
 *
 * Ebony board specific routines
 *
 * Matt Porter <mporter@mvista.com>
 *
 * Copyright 2002 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/config.h>
#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/reboot.h>
#include <linux/pci.h>
#include <linux/kdev_t.h>
#include <linux/types.h>
#include <linux/major.h>
#include <linux/blk.h>
#include <linux/console.h>
#include <linux/delay.h>
#include <linux/ide.h>
#include <linux/irq.h>
#include <linux/seq_file.h>
#include <linux/serial.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/time.h>
#ifdef CONFIG_PPC_RTC
#include <asm/todc.h>
#endif /* CONFIG_PPC_RTC */
#include <asm/bootinfo.h>
#include <asm/ppc4xx_pic.h>

#include "ds1672_rtc.h"
#include <system_interrupt_code.h>

extern void abort(void);

/* Global Variables */
unsigned char __res[sizeof (bd_t)];
struct semaphore i2c_lock;
unsigned char i2c_controller_slave_addr;
int i2c_controller_fast_mode;


static void __init
ebony_calibrate_decr(void)
{
	unsigned int freq;

	/*
	 * Determine system clock speed
	 *
	 * If we are on Rev. B silicon, then use
	 * default external system clock.  If we are
	 * on Rev. C silicon then errata forces us to
	 * use the internal clock.
	 */
	switch (PVR_REV(mfspr(PVR))) {
		case PVR_REV(PVR_440GP_RB):
			freq = EBONY_440GP_RB_SYSCLK;
			break;
		case PVR_REV(PVR_440GP_RC1):
		default:
			freq = EBONY_440GP_RC_SYSCLK;
			break;
	}
	
	tb_ticks_per_jiffy = freq / HZ;
	tb_to_us = mulhwu_scale_factor(freq, 1000000);

	/* Set the time base to zero */
	mtspr(SPRN_TBWL, 0);
	mtspr(SPRN_TBWU, 0);

	/* Clear any pending timer interrupts */
	mtspr(SPRN_TSR, TSR_ENW | TSR_WIS | TSR_DIS | TSR_FIS);

	/* Enable decrementer interrupt */
	mtspr(SPRN_TCR, TCR_DIE);
}

static int
ebony_show_cpuinfo(struct seq_file *m)
{
	seq_printf(m, "vendor\t\t: Topspin Communications\n");
	seq_printf(m, "machine\t\t: 120 Fabric Switch Card\n");

	return 0;
}

/* Some IRQs unique to Walnut.
 * Used by the generic 405 PCI setup functions in ppc4xx_pci.c
 */
static inline int
ebony_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{
        enum {
                NONE = -1,
                IRQ0 = 23,
                IRQ1 = 24,
                IRQ2 = 25,
                IRQ3 = 26,
                IRQ4 = 27,
                IRQ5 = 28,
                IRQ6 = 29
        };

	static char pci_irq_table[][4] =
	    /*
	     *      PCI IDSEL/INTPIN->INTLINE
	     *      A       B       C       D
	     */
	{
		{IRQ0, NONE, NONE, NONE}, /* IDSEL 1: Tavor                 */
		{IRQ1, IRQ2, NONE, NONE}, /* IDSEL 2: 82546 Chip A          */
		{IRQ3, IRQ4, NONE, NONE}, /* IDSEL 3: 82546 Chip B          */
		{NONE, NONE, NONE, NONE}, /* IDSEL 4                        */
	};

	const long min_idsel = 1, max_idsel = 4, irqs_per_slot = 4;
	return PCI_IRQ_TABLE_LOOKUP;
};

#define PCIX_READL(offset) \
	(readl((u32) pcix_reg_base + offset))
#define PCIX_READW(offset) \
	(readw((u32) pcix_reg_base + offset))
#define PCIX_WRITEL(value, offset) \
	(writel(value, (u32)pcix_reg_base+offset))
#define PCIX_WRITEW(value, offset) \
	(writew(value, (u32)pcix_reg_base+offset))

#ifdef NOTYET
/*
 * FIXME: This is only here to "make it work".  This will move
 * to a ibm_pcix.c which will contain a generic IBM PCIX bridge
 * configuration library. -Matt
 */
static void __init
ebony_setup_pcix(void)
{
	void *pcix_reg_base;

	pcix_reg_base = ioremap64(PCIX0_REG_BASE, PCIX0_REG_SIZE);

        /* Set up Bridge Options 1 register like openbios */
        {
          u32 brdgopt1;

          brdgopt1 = PCIX_READL(PCIX0_BRDGOPT1);
          brdgopt1 &= ~(0xffff0006);
          brdgopt1 |= 0x10000078;
#if 1 /* From ts-igr-fc.c.  Scott */
          brdgopt1 &= ~(0x00000060);  /* Set PLB pri to 0 */
          brdgopt1 |= 0x10000000;
#else
          brdgopt1 &= ~PCIX_BRDGOPT1_PLREQ_MASK;
          brdgopt1 |= PCIX_BRDGOPT1_PLREQ_0;
#endif
     
          PCIX_WRITEL(brdgopt1, PCIX0_BRDGOPT1);
        }

        /* Set bus 0, dev num 0 for PCI-X split transactions */
        {
          u32 pcixsts;

          pcixsts = PCIX_READL(PCIX0_STS);
          pcixsts &= ~0xfff8;
          PCIX_WRITEL(pcixsts, PCIX0_STS);
        }

        /* Enable bus mastering and memory transactions */
        {
          u16 command;

          command = PCIX_READW(PCIX0_COMMAND);
          command |= PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY; 
          PCIX_WRITEW(command, PCIX0_COMMAND);
        }

	/* Disable all windows */
	PCIX_WRITEL(0, PCIX0_POM0SA);
	PCIX_WRITEL(0, PCIX0_POM1SA);
	PCIX_WRITEL(0, PCIX0_POM2SA);
	PCIX_WRITEL(0, PCIX0_PIM0SA);
	PCIX_WRITEL(0, PCIX0_PIM1SA);
	PCIX_WRITEL(0, PCIX0_PIM2SA);
	
	/* Setup 2GB PLB->PCI outbound mem window (3_8000_0000->0_8000_0000) */
	PCIX_WRITEL(0x00000003, PCIX0_POM0LAH);
	PCIX_WRITEL(0x80000000, PCIX0_POM0LAL);
	PCIX_WRITEL(0x00000000, PCIX0_POM0PCIAH);
	PCIX_WRITEL(0x80000000, PCIX0_POM0PCIAL);
	PCIX_WRITEL(0x80000001, PCIX0_POM0SA);

	/* Setup 2GB PCI->PLB inbound memory window at 0, enable MSIs */
	PCIX_WRITEL(0x00000000, PCIX0_PIM0LAH);
	PCIX_WRITEL(0x00000000, PCIX0_PIM0LAL);
	PCIX_WRITEL(0xe0000007, PCIX0_PIM0SA);

	eieio();
}

static void __init
ebony_setup_hose(void)
{
	struct pci_controller *hose;

	/* Configure windows on the PCI-X host bridge */
	ebony_setup_pcix();

	hose = pcibios_alloc_controller();

	if (!hose)
		return;

	hose->first_busno = 0;
	hose->last_busno = 0xff;

	hose->pci_mem_offset = EBONY_PCI_MEM_OFFSET;

	pci_init_resource(&hose->io_resource,
			EBONY_PCI_LOWER_IO,
			EBONY_PCI_UPPER_IO,
			IORESOURCE_IO,
			"PCI host bridge");

	pci_init_resource(&hose->mem_resources[0],
			EBONY_PCI_LOWER_MEM,
			EBONY_PCI_UPPER_MEM,
			IORESOURCE_MEM,
			"PCI host bridge");

	hose->io_space.start = EBONY_PCI_LOWER_IO;
	hose->io_space.end = EBONY_PCI_UPPER_IO;
	hose->mem_space.start = EBONY_PCI_LOWER_MEM;
	hose->mem_space.end = EBONY_PCI_UPPER_MEM;
	isa_io_base =
		(unsigned long)ioremap64(EBONY_PCI_IO_BASE, EBONY_PCI_IO_SIZE);
	hose->io_base_virt = (void *)isa_io_base;

	setup_indirect_pci(hose,
			EBONY_PCI_CFGA_PLB32,
			EBONY_PCI_CFGD_PLB32);
	ppc_md.set_cfg_type = 1;

	hose->last_busno = pciauto_bus_scan(hose, hose->first_busno);

	ppc_md.pci_swizzle = common_swizzle;
	ppc_md.pci_map_irq = ebony_map_irq;
}
#endif /* NOTYET */

#ifdef CONFIG_PPC_RTC
TODC_ALLOC();
#endif /* CONFIG_PPC_RTC */

static void __init
ebony_setup_arch(void)
{
	struct serial_struct serial_req;

#if !defined(CONFIG_BDI_SWITCH)
	/*
	 * The Abatron BDI JTAG debugger does not tolerate others
	 * mucking with the debug registers.
	 */
        mtspr(SPRN_DBCR0, (DBCR0_TDE | DBCR0_IDM));
#endif

#ifdef CONFIG_PPC_RTC
	/* Setup TODC access */
	TODC_INIT(TODC_TYPE_DS1743,
			0,
			0,
			ioremap64(EBONY_RTC_ADDR, EBONY_RTC_SIZE),
			8);
#endif /* CONFIG_PPC_RTC */

	/* init to some ~sane value until calibrate_delay() runs */
        loops_per_jiffy = 50000000/HZ;

#ifdef NOTYET
    /* The 120 Fabric SC doesn't have any PCI devices */
	/* Setup PCI host bridge */
	ebony_setup_hose();
#endif

#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start)
		ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0); /* /dev/ram */
	else
#endif
#ifdef CONFIG_ROOT_NFS
		ROOT_DEV = to_kdev_t(0x00FF);	/* /dev/nfs pseudo device */
#else
		ROOT_DEV = to_kdev_t(0x1601);	/* /dev/hdc1 */
#endif

#ifdef CONFIG_DUMMY_CONSOLE
	conswitchp = &dummy_con;
#endif

	/* Setup ioremapped serial port access */
	memset(&serial_req, 0, sizeof(serial_req));
	serial_req.line = 0;
	serial_req.baud_base = BASE_BAUD;
	serial_req.port = 0;
	serial_req.irq = 0;
	serial_req.flags = ASYNC_BOOT_AUTOCONF | ASYNC_SKIP_TEST;
	serial_req.io_type = SERIAL_IO_MEM;
	serial_req.iomem_base = ioremap64(PPC440GP_UART0_ADDR, 8);
	serial_req.iomem_reg_shift = 0;

	if (early_serial_setup(&serial_req) != 0) {
		printk("Early serial init of port 0 failed\n");
	}

	/* Assume early_serial_setup() doesn't modify serial_req */
	serial_req.line = 1;
	serial_req.port = 1;
	serial_req.irq = 1; 
	serial_req.iomem_base = ioremap64(PPC440GP_UART1_ADDR, 8);

	if (early_serial_setup(&serial_req) != 0) {
		printk("Early serial init of port 1 failed\n");
	}

	/* Identify the system */
	printk("Topspin 120 Fabric Switch Card\n");
}

static void
ts120_restart(char *cmd)
{
    char *sram_sig = (char *)PPC440_SRAM_VIRT_BASE;

    strcpy(sram_sig, SOFTWARE_RESET);

    /* disable DMA master in the FPGA here */
    ts120_disable_dma();
    printk("Disabled DMA Master in the FPGA\n");

    /* keep anafa in reset */
    ts120_anafa_reset();
    printk("Kept Anafa in reset\n");
    udelay(500);

    init_ts_watchdog();
    printk("Activated the hardware watchdog\n");
    udelay(500);

	__cli();
	abort();
}

static void
ebony_power_off(void)
{
	__cli();
	for(;;);
}

static void
ebony_halt(void)
{
	__cli();
	for(;;);
}

/*
 * Read the 440GP memory controller to get size of system memory.
 */
static unsigned long __init
ebony_find_end_of_memory(void)
{
	u32 i, bank_config;
	u32 mem_size = 0;

	for (i=0; i<4; i++)
	{
		switch (i)
		{
			case 0:
				mtdcr(DCRN_SDRAM0_CFGADDR, SDRAM0_B0CR);
				break;
			case 1:
				mtdcr(DCRN_SDRAM0_CFGADDR, SDRAM0_B1CR);
				break;
			case 2:
				mtdcr(DCRN_SDRAM0_CFGADDR, SDRAM0_B2CR);
				break;
			case 3:
				mtdcr(DCRN_SDRAM0_CFGADDR, SDRAM0_B3CR);
				break;
		}

		bank_config = mfdcr(DCRN_SDRAM0_CFGDATA);

		if (!(bank_config & SDRAM_CONFIG_BANK_ENABLE))
			continue;
		switch (SDRAM_CONFIG_BANK_SIZE(bank_config))
		{
			case SDRAM_CONFIG_SIZE_8M:
				mem_size += PPC440_MEM_SIZE_8M;
				break;
			case SDRAM_CONFIG_SIZE_16M:
				mem_size += PPC440_MEM_SIZE_16M;
				break;
			case SDRAM_CONFIG_SIZE_32M:
				mem_size += PPC440_MEM_SIZE_32M;
				break;
			case SDRAM_CONFIG_SIZE_64M:
				mem_size += PPC440_MEM_SIZE_64M;
				break;
			case SDRAM_CONFIG_SIZE_128M:
				mem_size += PPC440_MEM_SIZE_128M;
				break;
			case SDRAM_CONFIG_SIZE_256M:
				mem_size += PPC440_MEM_SIZE_256M;
				break;
			case SDRAM_CONFIG_SIZE_512M:
				mem_size += PPC440_MEM_SIZE_512M;
				break;
		}
	}
	return mem_size;
}

static void __init
ebony_init_irq(void)
{
	int i;

	ppc4xx_pic_init();

	for (i = 0; i < NR_IRQS; i++)
		irq_desc[i].handler = ppc4xx_pic;
}

#ifdef CONFIG_SERIAL_TEXT_DEBUG
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>

static struct serial_state rs_table[RS_TABLE_SIZE] = {
	SERIAL_PORT_DFNS	/* Defined in <asm/serial.h> */
};

static void
ebony_progress(char *s, unsigned short hex)
{
	volatile char c;
	volatile unsigned long com_port;
	u16 shift;

	com_port = (unsigned long)rs_table[0].iomem_base;
	shift = rs_table[0].iomem_reg_shift;

	while ((c = *s++) != 0) {
		while ((*((volatile unsigned char *)com_port +
				(UART_LSR << shift)) & UART_LSR_THRE) == 0)
			;
		*(volatile unsigned char *)com_port = c;

	}

	/* Send LF/CR to pretty up output */
	while ((*((volatile unsigned char *)com_port +
		(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		;
	*(volatile unsigned char *)com_port = '\r';
	while ((*((volatile unsigned char *)com_port +
		(UART_LSR << shift)) & UART_LSR_THRE) == 0)
		;
	*(volatile unsigned char *)com_port = '\n';
}
#endif /* CONFIG_SERIAL_TEXT_DEBUG */

void __init platform_init(unsigned long r3, unsigned long r4,
		unsigned long r5, unsigned long r6, unsigned long r7)
{
	parse_bootinfo((struct bi_record *) (r3 + KERNELBASE));
	if (r3) {
		memcpy((void *) __res, (void *) (r3 + KERNELBASE), sizeof (bd_t));
        }

#if defined(CONFIG_BLK_DEV_INITRD)
	/*
	 * If the init RAM disk has been configured in, and there's a valid
	 * starting address for it, set it up.
	 */
	if (r4) {
		initrd_start = r4 + KERNELBASE;
		initrd_end = r5 + KERNELBASE;
	}
#endif				/* CONFIG_BLK_DEV_INITRD */

	/* Copy the kernel command line arguments to a safe place. */

	if (r6) {
		*(char *) (r7 + KERNELBASE) = 0;
		strcpy(cmd_line, (char *) (r6 + KERNELBASE));
	}

	ppc_md.setup_arch = ebony_setup_arch;
	ppc_md.show_cpuinfo = ebony_show_cpuinfo;
	ppc_md.init_IRQ = ebony_init_irq;
	ppc_md.get_irq = NULL;		/* Set in ppc4xx_pic_init() */

	ppc_md.find_end_of_memory = ebony_find_end_of_memory;
	ppc_md.setup_io_mappings = NULL;

	ppc_md.restart = ts120_restart;
	ppc_md.power_off = ebony_power_off;
	ppc_md.halt = ebony_halt;

	ppc_md.calibrate_decr = ebony_calibrate_decr;

#ifdef CONFIG_PPC_RTC
	ppc_md.time_init = todc_time_init;
	ppc_md.set_rtc_time = todc_set_rtc_time;
	ppc_md.get_rtc_time = todc_get_rtc_time;

	ppc_md.nvram_read_val = todc_direct_read_val;
	ppc_md.nvram_write_val = todc_direct_write_val;
#endif /* CONFIG_PPC_RTC */

    init_MUTEX(&i2c_lock);
    i2c_controller_slave_addr = 0;            /* not in slave mode */
    i2c_controller_fast_mode = 0;             /* slow mode 100 KHz */

    ppc_md.time_init = ds1672_time_init;
    ppc_md.set_rtc_time = ds1672_set_rtc_time;
    ppc_md.get_rtc_time = ds1672_get_rtc_time;

    ppc_md.nvram_read_val = NULL;
    ppc_md.nvram_write_val = NULL;


#ifdef CONFIG_SERIAL_TEXT_DEBUG
	ppc_md.progress = ebony_progress;
#endif /* CONFIG_SERIAL_TEXT_DEBUG */

  /* Use external clock for serial ports */
  mtdcr(DCRN_CPC0_CR0,
              mfdcr(DCRN_CPC0_CR0) | CR0_U0EC | CR0_U1EC);

}

